/*
 * mpl-udev.c
 *
 * by Alexander Zhirov (alexander@zhirov.kz)
 * Telegram @alexanderzhirov
 */

#include "mpl-udev.h"
#include "mpl-core.h"

#include <unistd.h>
#include <sys/wait.h>

void test(struct udev_device *dev)
{
    const char *action = udev_device_get_action(dev);
    // printf("action: %s\n", action);
    // const char *devpath = udev_device_get_devpath(dev);
    // printf("devpath: %s\n", devpath);

    // const char *subsystem = udev_device_get_subsystem(dev);
    // printf("subsystem: %s\n", subsystem);

    const char *devtype = udev_device_get_devtype(dev);
    // printf("devtype: %s\n", devtype);

    const char *driver = udev_device_get_driver(dev);
    // printf("driver: %s\n", driver);

    // const char *sysattr_value = udev_device_get_sysattr_value(dev, "bInterfaceNumber");
    // printf("Номер порта: %s\n", sysattr_value);

    const char *syspath = udev_device_get_syspath(dev);
    // printf("syspath: %s\n", syspath);

    // const char *sysname = udev_device_get_sysname(dev);
    // printf("sysname: %s\n", sysname);

    // const char *sysnum = udev_device_get_sysnum(dev);
    // printf("sysnum: %s\n", sysnum);

    const char *devnode = udev_device_get_devnode(dev);
    // printf("devnode: %s\n", devnode);

    // int is_initialized = udev_device_get_is_initialized(dev);
    // printf("is_initialized: %d\n", is_initialized);

    printf("action: %s\tdevtype: %s\tdriver: %s\tsyspath: %s\tdevnode: %s\n", action, devtype, driver, syspath, devnode);
}

int mpl_udev_init(MPL_UDEV *udev)
{
    udev->context = udev_new();
    if (udev->context == NULL)
    {
        MPLOG(LOG_ERR, "error creating udev context");
        return 1;
    }

    udev->monitor = udev_monitor_new_from_netlink(udev->context, "udev");
    if (udev->monitor == NULL)
    {
        MPLOG(LOG_ERR, "error creating udev monitor");
        mpl_udev_free(udev);
        return 1;
    }

    udev_monitor_filter_add_match_subsystem_devtype(udev->monitor, "usb", NULL);
    udev_monitor_enable_receiving(udev->monitor);

    udev->fd = udev_monitor_get_fd(udev->monitor);
    if (udev->fd < 0)
    {
        MPLOG(LOG_ERR, "error creating udev monitor descriptor");
        mpl_udev_free(udev);
        return 1;
    }

    return 0;
}

// action: bind    devtype: usb_interface  driver: option  syspath: /sys/devices/pci0000:00/0000:00:14.0/usb3/3-3/3-3.3/3-3.3:1.1
// action: unbind  devtype: usb_interface  driver: (null)  syspath: /sys/devices/pci0000:00/0000:00:14.0/usb3/3-3/3-3.3/3-3.3:1.1

static int mpl_device_is_interface(struct udev_device *dev)
{
    return strcmp(udev_device_get_devtype(dev), "usb_interface") == 0;
}

static int mpl_device_is_modem(struct udev_device *dev)
{
    const char *driver = udev_device_get_driver(dev);
    const char *sysattr = udev_device_get_sysattr_value(dev, "bInterfaceNumber");

    if (!driver || !sysattr)
        return 0;

    return mpl_device_is_interface(dev) && strcmp(driver, "option") == 0 && sysattr;
}

static int mpl_connect_modem(struct udev_device *dev)
{
    const char *action = udev_device_get_property_value(dev, "ACTION");

    return strcmp(action, "bind") == 0 && mpl_device_is_modem(dev);
}

static int mpl_disconnect_modem(struct udev_device *dev)
{
    const char *action = udev_device_get_property_value(dev, "ACTION");

    return strcmp(action, "unbind") == 0 && mpl_device_is_interface(dev);
}

static MPL_MODEM *mpl_get_modem_info(struct udev_device *dev)
{
    MPL_MODEM *modem = (MPL_MODEM *)calloc(1, sizeof(MPL_MODEM));

    if (modem == NULL)
    {
        MPLOG(LOG_ERR, "error creating modem");
        return NULL;
    }

    modem->port = atoi(udev_device_get_sysattr_value(dev, "bInterfaceNumber"));
    modem->path = mpl_get_copy_string(udev_device_get_syspath(dev));
    modem->action = strcmp(udev_device_get_property_value(dev, "ACTION"), "bind") == 0;
    
    return modem;
}

static void mpl_free_modem(MPL_MODEM *modem)
{
    if (modem->path)
        free(modem->path);
    free(modem);
}

int mpl_udev_check_devices(MPL_UDEV *udev)
{
    struct udev_device *dev;
    struct udev_enumerate *enumerate;
    struct udev_list_entry *devices, *dev_list_entry;

    MPL_MODEM *modem = NULL;

    enumerate = udev_enumerate_new(udev->context);

    if (enumerate == NULL)
    {
        MPLOG(LOG_ERR, "error creating udev enumerate");
        return 1;
    }

    udev_enumerate_add_match_subsystem(enumerate, "usb");
    udev_enumerate_scan_devices(enumerate);

    devices = udev_enumerate_get_list_entry(enumerate);

    if (devices == NULL)
    {
        MPLOG(LOG_ERR, "error get udev devices");
        return 1;
    }

    udev_list_entry_foreach(dev_list_entry, devices)
    {
        const char *syspath = udev_list_entry_get_name(dev_list_entry);
        dev = udev_device_new_from_syspath(udev->context, syspath);

        if (!mpl_device_is_modem(dev))
            continue;

        modem = mpl_get_modem_info(dev);

        device_path(modem);

        mpl_free_modem(modem);
        udev_device_unref(dev);
    }

    udev_enumerate_unref(enumerate);

    return 0;
}

void mpl_udev_free(MPL_UDEV *udev)
{
    if (udev->monitor)
    {
        udev_monitor_unref(udev->monitor);
        udev->monitor = NULL;
    }
    if (udev->context)
    {
        udev_unref(udev->context);
        udev->context = NULL;
    }
}

void mpl_udev_loop(MPL_UDEV *udev, MPL_CANCELLABLE *cancellable)
{
    MPL_MODEM *modem = NULL;
    int status = -1;
    pid_t pid = -1;

    int stop = 0;

    while (!stop)
    {
        fd_set fds;
        FD_ZERO(&fds);
        FD_SET(udev->fd, &fds);

        if (select(udev->fd + 1, &fds, NULL, NULL, NULL) <= 0)
        {
            if (cancellable->exit)
                return;

            MPLOG(LOG_ERR, "file selection error");
            break;
        }

        struct udev_device *dev = udev_monitor_receive_device(udev->monitor);
        if (dev)
        {
            // test(dev);

            if (mpl_connect_modem(dev))
            {
                modem = mpl_get_modem_info(dev);

                pid = fork();
                if (pid == 0)
                {
                    stop = 1;
                }
                else
                    wait(&status);
            }
            else if (mpl_disconnect_modem(dev))
            {
                // unmount ports
            }
            
            udev_device_unref(dev);
        }
    }

    mpl_udev_free(udev);

    if (cancellable->exit)
        return;

    // printf("path: %s\t port: %d\t action: %d\n", modem->path, modem->port, modem->action);

    device_path(modem);
    mpl_free_modem(modem);
}
